from os import path , getcwd , stat

class arFile:        
    def GetCurrentDirectory(self, AsFolder :str = "") -> str:
        LsCurrPath = getcwd()
        if (AsFolder != ""):
            LsCurrPath = path.join(LsCurrPath, AsFolder)
        return LsCurrPath

    def IsFileExist(self , AsFilePath :str) -> bool:
        return path.exists(AsFilePath)

    def IsEmptyFile(self , AsFilePath :str) -> bool:
        return (stat(AsFilePath).st_size == 0)

    def CreateFile(self, AsFilePath :str, AsText :str = "") -> bool:
        try:
            file = open(AsFilePath, "w")
            file.write(AsText)                                               
            return True
        except Exception as e:
            print(f"Error creating file : {e}")
            return False      



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from AbdallahRadwanLib.arUtilityConst import ardbSettings

"""
    1. Create new engine
    2. create new declarative base
    3. base with metadata and assign engine
    4. create new session maker
    5. Inherit from new object from session maker
"""

# Database configuration
Base = declarative_base()

class arSQLAlchemyManager:
    __instance = None
    @staticmethod
    def CreateInstance(dbSetting :ardbSettings):
        if (arSQLAlchemyManager.__instance == None):
            arSQLAlchemyManager.__instance = arSQLAlchemyManager(dbSetting)
        return arSQLAlchemyManager.__instance
    
    def __init__(self, dbSetting :ardbSettings):
        """Initialize the database manager with the provided database URL."""        
        EnableEcho,dbType = dbSetting.EnableEcho, dbSetting.dbType.upper()
        if (dbType == "ORACLE"):        
            database_url = f"oracle+cx_oracle://{dbSetting.dbUser}:{dbSetting.dbPass}@{dbSetting.dbHost}:{dbSetting.dbPort}/{dbSetting.dbService}"
        elif (dbType == "SQLITE"):        
            database_url = f"sqlite:///{dbSetting.dbPath}\\Database\\SQLite\\{dbSetting.dbName}.db"
        elif (dbType == "MONGODB"):        
            database_url = f"mongodb://{dbSetting.dbHost}:{dbSetting.dbPort}/"          
        else:
            database_url = ""  
        print(f"Welcome to SqlAlchemy : Database [{dbType}] => Alias [{dbSetting.dbAlias}] => URL [{database_url}]")
        # إعداد الاتصال بقاعدة البيانات
        self.engine = create_engine(database_url, echo=EnableEcho)
        self.Base = declarative_base()
        self.SessionMaker = sessionmaker(bind=self.engine)        
        self.create_tables()  # Create tables if they don't exist  

    def get_base(self):
        return self.Base
    
    def create_tables(self):
        """Create the database tables."""
        # إنشاء الجداول
        self.Base.metadata.create_all(bind=self.engine)

    def get_session(self):
        # إنشاء جلسة للتفاعل مع قاعدة البيانات
        """Return a new session."""        
        return self.SessionMaker()


from AbdallahRadwanLib.arUtilityFile import arFile
from AbdallahRadwanLib.arUtilityEnum import arEnumConfigType
from AbdallahRadwanLib.arUtilityConst import ardbResult , ardbSettings
from AbdallahRadwanLib.arUtilitySettings import BasicSettings , VariableSettings

class arConfig:      
    __ConfigFileName, __ConfigFilePath = "Settings.json" , ""     
    __ConfigJsonData = {}
    __DbSettings = None

    __instance = None

    @staticmethod
    def CreateInstance():
        if (arConfig.__instance == None):
            arConfig.__instance = arConfig()
        return arConfig.__instance
    
    def __init__(self) -> None:    
        # print(f"Create Config : ")    
        self.CreateSettingFile()        
        self.__SetDbSettings()          
        

    def __SetConfigFilePath(self):
        fileUtility = arFile()
        self.__ConfigFilePath = fileUtility.GetCurrentDirectory(self.__ConfigFileName)        

    def __AddNewSetting(self, dictBasicData :dict, subKey :str, subValue :object):
        IsExistItem = False
        try:        
            for mainKey,mainValue in dictBasicData.items():
                if (mainKey == subKey):                
                    if (subValue != None):
                        IsExistItem = True                    
                        break                    
            if (not IsExistItem):
                dictBasicData.update( {subKey : subValue} )
        except Exception as e:
            print(f"__AddNewSetting.Exception : {e}")

    def __GetBasicData(self) -> dict:        
        import json
        dictBasicData    = BasicSettings
        dictVariableData = VariableSettings
        fileUtility = arFile()
        if (fileUtility.IsFileExist(self.__ConfigFilePath)):
            if (not fileUtility.IsEmptyFile(self.__ConfigFilePath)):
              file = open(self.__ConfigFilePath , "+r")            
              dictBasicData = json.load(file)              
              file.seek(0)
        else:            
            fileUtility.CreateFile(self.__ConfigFilePath)

        return dictBasicData, dictVariableData
    
    def __SetConfigJsonData(self):        
        result = ""
        try:                                    
            self.__ConfigJsonData, dictVariableData  = self.__GetBasicData()                                                            
            for Key,Value in dictVariableData.items():
                self.__AddNewSetting(self.__ConfigJsonData,Key,Value)            
            # Serializing json            
            import json
            JsonData = json.dumps(self.__ConfigJsonData, indent=4)
            fileUtility = arFile() 
            fileUtility.CreateFile(self.__ConfigFilePath, JsonData)
        except Exception as e:            
            result =""
            errorMsg = f"__SetConfigJsonData.Exception : {e}"
            print(errorMsg)

    def __SetDbSettings(self) -> None:        
        self.__DbSettings = ardbSettings()       
        self.__DbSettings.dbAlias = self.GetConfigValue(arEnumConfigType.defConnStr)
        self.__DbSettings.dbFetchCount =  self.GetConfigValue(arEnumConfigType.fetchCount)        
        if (len(self.__DbSettings.dbAlias) > 0):            
            connStrs = self.GetConfigValue(arEnumConfigType.connStrs)            
            try:                
                for connStr in connStrs :                    
                    connStrName = connStr.get("alias")                            
                    if (connStrName == self.__DbSettings.dbAlias):                           
                        self.__DbSettings.dbType        = connStr.get("type")
                        self.__DbSettings.dbName        = connStr.get("name")
                        self.__DbSettings.dbUser        = connStr.get("userName")
                        self.__DbSettings.dbPass        = connStr.get("password")
                        self.__DbSettings.dbHost        = connStr.get("host")                        
                        self.__DbSettings.dbPort        = connStr.get("port")
                        self.__DbSettings.dbService     = connStr.get("service")
                        break
            except Exception as e:
                errorMsg = f"__SetDbSettings.Exception : {e}"
                print(errorMsg)
        
    def CreateSettingFile(self):        
        self.__SetConfigFilePath()                
        self.__SetConfigJsonData()  

    def GetConfig(self, key :str) -> object:
        try:
            result = self.__ConfigJsonData.get(key)
        except Exception as e:            
            errorMsg = f"GetConfig.Exception : {e}"
            result = ""            
        return result
    
    def GetConfigValue(self, Enumkey :arEnumConfigType) -> object:
        # print(f"Enumkey : {Enumkey.name} : {Enumkey.value}")
        return self.GetConfig(Enumkey.value)
        
    def GetDbSettings(self) -> object:
        return self.__DbSettings
    
    @classmethod
    def ClassMethodReturnInstance(cls):
        return cls()        

config = arConfig.CreateInstance()


# to get a string like this run: openssl rand -hex 32
jwt_SecretKey = "0b0de311e50faaf5fc8181547644b83fa55caa1c32d025ea0ed72feeccb04be5"       
jwt_Algorithm = "HS256"
jwt_ExpireHours = 24

class ardbSettings:
    dbType       :str = "" 
    dbAlias      :str = "" 
    dbName       :str = ""
    dbUser       :str = ""
    dbPass       :str = ""
    dbHost       :str = ""
    dbService    :str = ""
    dbPath       :str = ""
    dbPort       :int = 0
    dbFetchCount :int = 0
    EnableEcho   :bool = False

class ardbResult:
    connected :bool=False
    isSuccess :bool=False
    message   :str=""
    columns   :dict={}
    rows      :object=None
    rowcount  :int=0
    data      :object=None




import jwt 
from cryptography.fernet import Fernet
from typing import Tuple
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from AbdallahRadwanLib.arUtilityEnum import arEnumEncryptionType
from AbdallahRadwanLib.arUtilityConst import jwt_SecretKey, jwt_Algorithm

class arEncryption:
    def GenerateToken(self, payload :dict) -> str:
        try:        
            token = jwt.encode(payload , jwt_SecretKey, jwt_Algorithm)
        except Exception as e:
            token = ""
            errorMessage = f"Error : {e}"
        return token   

    def ReadToken(self, token :str) -> dict:
        payload = {}
        try:                           
            payload = jwt.decode(token , jwt_SecretKey, jwt_Algorithm, options={"verify_aud": False})
        except Exception as e:
            payload = {}
            errorMessage = f"Exception : {e}"
        return payload   

    def __EncHash(self,text :str) -> str:        
        try:
            value = str(hash(text))
        except:
            value = ""            
        return value
            
    def __EncFernet(self,text :str) -> Tuple[str,str]:#using salt 
        try:            
            value,key = "",""
            key = Fernet.generate_key()  
            fernet = Fernet(key)
            value = fernet.encrypt(text.encode()).decode()                        
            key = key.decode()            
        except:
            value,key = "",""
        return value,key

    def __DecFernet(self,text :str, key :str) -> str:
        try:
            value = ""
            value = Fernet(key).decrypt(text).decode()
        except:
            value = ""
        return value    
    
    def __EncAES(self,text :str) -> Tuple[str,str,str,str]:#using salt         
        try:    
            value,key,tag,nonce = "","","",""
            # AES requires a key of 16, 24, or 32 bytes        
            key = get_random_bytes(16)  # Generate a random 16-byte key 
            cipher = AES.new(key, AES.MODE_EAX)
            nonce = cipher.nonce
            value, tag = cipher.encrypt_and_digest(text.encode("ascii"))        
            value,key,tag,nonce = value.hex(), key.hex(), tag.hex(), nonce.hex()
        except Exception as e:
            error = f"EncAES : {e}"
            value,key,tag,nonce = "","","",""
        return value,key,tag,nonce

    def __DecAES(self,text :str, key :str, tag :str, nonce :str) -> str:        
        try:
            value = ""
            text,key,tag,nonce = bytes.fromhex(text),bytes.fromhex(key),bytes.fromhex(tag),bytes.fromhex(nonce)

            cipher = AES.new(key, AES.MODE_EAX, nonce=nonce)
            value = cipher.decrypt(text)   
            cipher.verify(tag)   
            value = value.decode("ascii")                 
        except Exception as e:
            error = f"DecAES : {e}"
            value = ""
        return value   
    
    def encodeText(self,text :str, encType: arEnumEncryptionType) -> Tuple[str,str,str,str]:        
        value,key,tag,nonce = "","","",""
        match encType: 
            case arEnumEncryptionType.etFernet:
                value,key = self.__EncFernet(text)
            case arEnumEncryptionType.etAES:
                value,key,tag,nonce = self.__EncAES(text)                
        return value,key,tag,nonce
            
    def decodeText(self,text :str, key :str, tag :str, nonce :str, encType: arEnumEncryptionType) -> str:
        value = ""
        match encType: 
            case arEnumEncryptionType.etFernet:
                value = self.__DecFernet(text, key)
            case arEnumEncryptionType.etAES:
                value = self.__DecAES(text, key, tag, nonce)                    
        return value



from datetime import datetime , timedelta
from AbdallahRadwanLib.arUtilityEncryption import arEncryption, arEnumEncryptionType
from AbdallahRadwanLib.arUtilityConst import jwt_ExpireHours
from typing import Tuple
from uuid import uuid4

class arUtility:    
    def GetRandomNumber(self, AiFrom :int=1, AiTo :int=100) -> int:
        import random
        return random.randrange(start=AiFrom, stop=AiTo)
    
    def GetCurrentDatetime(self) -> datetime:        
        return datetime.now()
    
    def GetCurrentTimestamp(self) -> float:        
        return datetime.now().timestamp()
    
    def GetDatefromTimestamp(self, stamp :float) -> datetime:
        return datetime.fromtimestamp(stamp)
    
    def GetTimestampfromDate(self, dt :datetime) -> float:        
        return dt.timestamp()
        
    def EncryptDate(self, dt) -> str:
        Enc = arEncryption()
        return Enc.encodeText(str(dt), arEnumEncryptionType.etAES)        
        
    def GetTimestampCurDateAndExpired(self) -> Tuple[str,str]:
        dt = self.GetCurrentDatetime()                
        ex = dt + timedelta(hours=jwt_ExpireHours)                                
        createdAt,expireIn = str(int(self.GetTimestampfromDate(dt))), str(int(self.GetTimestampfromDate(ex)))
        return createdAt,expireIn
    
    def GetDateAndExpiredFromTimestamp(self,dt :str,ex :str) -> Tuple[datetime|datetime]:
        if ( (dt == None) | (ex == None) ):
              return None,None        
        dtValue,exValue = float(dt),float(ex)
        createdAt,expireIn = self.GetDatefromTimestamp(dtValue),self.GetDatefromTimestamp(exValue)        
        return createdAt,expireIn
    
    def IsExpiredDate(self, dt :str) ->bool :        
        if (dt == None):
            return True        
        try:                    
            dtValue = datetime.strptime(dt,"%Y-%m-%d %H:%M:%S")
            IsExpired = (dtValue < datetime.now())
        except:
            IsExpired = False
        return IsExpired

    def GetGUID(self, removeDash :bool) -> str:
        value = str(uuid4())
        if (removeDash):
            value = value.replace("-","")
        return value


BasicSettings = {                
                    "Version" : "1.0.0.0",    
                    "Language" : "Python",
                    "Framework" : "FastAPI",
                    "enviroments": [
                        "dev",
                        "test",
                        "prod"
                    ],                    
                    "connectionStrings": [
                                            {
                                                "type": "Oracle",
                                                "alias": "AbhaTest",
                                                "name": "",
                                                "userName": "RCAPH",
                                                "password": "RCAPH",
                                                "host": "AIPHDSVM",
                                                "port": 1521,
                                                "serviceName": "AIPHTEST"
                                            },
                                            {
                                                "type": "Oracle",
                                                "alias": "AbhaProduction",
                                                "name": "",
                                                "userName": "RCAPH",
                                                "password": "SCRC5DEVAPH18",
                                                "host": "AIPHDSVM",
                                                "port": 1521,
                                                "serviceName": "aiphdb"
                                            },
                                            {
                                                "type": "Oracle",
                                                "alias": "Abdallah_W",
                                                "name": "",
                                                "userName": "ABDALLAH",
                                                "password": "ABDALLAH",
                                                "host": "AIPHDSVM",
                                                "port": 1521,
                                                "serviceName": "AIPHTEST"
                                            },
                                            {
                                                "type": "Oracle",
                                                "alias": "Abdallah_H",
                                                "name": "",
                                                "userName": "ABDALLAH",
                                                "password": "ABDALLAH",
                                                "host": "Abdallah-Lat5540",
                                                "port": 1521,
                                                "serviceName": "XE"
                                            },
                                            {
                                                "type": "SQLite",
                                                "alias": "dbSQLite",
                                                "name": "hospital",
                                                "userName": "",
                                                "password": "",
                                                "host": "",
                                                "port": 0,
                                                "serviceName": ""
                                            },
                                            {
                                                "type": "MongoDB",
                                                "alias": "dbMongoDB",
                                                "name": "",
                                                "userName": "",
                                                "password": "",
                                                "host": "localhost",
                                                "port": 27017,
                                                "serviceName": ""
                                            },                                            
                                        ],                
                }

VariableSettings = {
                        "DefaultEnviroment": "dev",
                        "DefaultConnectionString" : "Abdallah_W",                        
                        "ServerIP" : "127.0.0.1",
                        "ServerPort" : 15500,
                        "FetchCount" : 10
                   }    


from enum import Enum

class arEnumConfigType(Enum):
    connStrs   = "connectionStrings"
    defEnv     = "DefaultEnviroment"
    defConnStr = "DefaultConnectionString"    
    serverIP   = "ServerIP"
    serverPort = "ServerPort"
    fetchCount = "FetchCount"

class arEnumEncryptionType(Enum):
    etHash   = "hash -> Hashing is a one-way functio"    
    etFernet = "Fernet -> symmetric encryption"
    etAES    = "AES -> Advanced Encryption Standard"   
    # etToken  = "Bearer Token" 

class arEnumFetchType(Enum):
    ftOne  = 1
    ftMany = 2
    ftAll  = 3        


from typing import Tuple
import cx_Oracle
from AbdallahRadwanLib.arUtilityConst import ardbResult , ardbSettings
from AbdallahRadwanLib.arUtilityEnum import arEnumFetchType


class arOracle:
    __oraUsername, __oraPassword, __oraHost, __oraServiceName,_ConnectionString = "","","","",""
    __oraPort , __fetchCount = 1521, 10

    def __init__(self, dbSettings :ardbSettings) -> None:    
        self.__oraUsername    = dbSettings.dbUser
        self.__oraPassword    = dbSettings.dbPass
        self.__oraHost        = dbSettings.dbHost
        self.__oraService     = dbSettings.dbService
        self.__oraPort        = dbSettings.dbPort
        self.__fetchCount     = dbSettings.dbFetchCount        
        self.__SetConnectionString()
        # print(f"Connection String : {self.__GetConnectionString()}\n")

    def __SetConnectionString(self) -> None:      
        self._ConnectionString = f"{self.__oraUsername}/{self.__oraPassword}@{self.__oraHost}:{self.__oraPort}/{self.__oraService}"

    def __GetConnectionString(self) -> str:                
        return self._ConnectionString
    
    def openConn(self) -> Tuple[object,bool]:
        try:            
            LoConn = cx_Oracle.connect(self.__GetConnectionString())            
        except:
            LoConn = None
        finally:
            LbConnected = (LoConn != None)
        return LoConn , LbConnected
    
    def closeConn(self , AoConn :object) -> None:
        try:
            if (AoConn != None):
                AoConn.close()
        except:
            pass
    
    def __getColumns(self, AoCursor :object) -> dict:
        return { cd[0]:i for i , cd in enumerate(AoCursor.description) }
    
    def testConnection(self):
        dbResult = ardbResult()
        try:
            LoConn , dbResult.connected = self.openConn()
        except Exception as e:
            dbResult.isSuccess,dbResult.message = False,f"Exception : {e}"
        finally:
            self.closeConn(LoConn)        
        return dbResult
                    
    def fetchData(self, AsSQL :str, AtpValue :tuple, AoFetchType :arEnumFetchType = arEnumFetchType.ftMany) -> ardbResult:
        dbResult = ardbResult()
        try:            
            if type(AtpValue) == str:                
                AtpValue = (AtpValue,) 
            dbResult.message = "Open database connection"
            LoConn , dbResult.connected = self.openConn()
            if (dbResult.connected):
                dbResult.message = "Open cursor"
                LoCursor = LoConn.cursor()                
                dbResult.message = "Execute Data"
                LoCursor.execute(AsSQL,AtpValue)
                dbResult.message = "Get list of column name"
                dbResult.columns = self.__getColumns(LoCursor)
                dbResult.message = "Fetch data ..."
                if (AoFetchType == arEnumFetchType.ftMany):
                    dbResult.rows = LoCursor.fetchmany(self.__fetchCount)
                else:
                    dbResult.rows = LoCursor.fetchall()
                dbResult.rowcount = LoCursor.rowcount
        except Exception as e:
            dbResult.isSuccess,dbResult.message,dbResult.columns,dbResult.rows,dbResult.rowcount = False,f"Exception : {e}",None,None,0
        finally:
            if (dbResult.connected):
                LoCursor.close()                
            self.closeConn(LoConn)        
        return dbResult
         
    def executeData(self, AoSQLList :list[str], AoValueList :list[tuple]) -> ardbResult:
        '''
            Cursor.execute(SQL as str)
            Cursor.execute(SQL as str, Row as tuple) # you can passing value by position or by name
            Cursor.executemany(SQL as str, Rows as List[tuple])
        '''
        dbResult = ardbResult()
        try:
            dbResult.message = "Open database connection"
            LoConn , dbResult.connected = self.openConn()
            if (dbResult.connected):
                dbResult.message = "Open cursor"
                LoCursor = LoConn.cursor()
                dbResult.message = "Execute Data"
                for LsSQL in AoSQLList:
                    LtpValue = AoValueList[AoSQLList.index(LsSQL)]
                    LoCursor.execute(LsSQL, (LtpValue))                    
                LoConn.commit()                
                dbResult.rowcount = LoCursor.rowcount
                dbResult.isSuccess = True
        except Exception as e:
            if (dbResult.connected):
                LoConn.rollback()
            dbResult.isSuccess,dbResult.message,dbResult.rowcount = False,f"Exception : {e}",0            
        finally:
            if (dbResult.connected):
                LoCursor.close()                
            self.closeConn(LoConn) 
        return dbResult
        

    def CallProc(self, AsName :str, AoParams :list) -> ardbResult:
        dbResult = ardbResult()
        try:            
            dbResult.message = "Open database connection"
            LoConn , dbResult.connected = self.openConn()
            if (dbResult.connected):
                dbResult.message = "Open cursor"
                LoCursor = LoConn.cursor()                
                dbResult.message = "Execute Data"
                LoCursor.callproc(AsName , AoParams)
                dbResult.isSuccess,dbResult.message = True, "Procedure Exceuted ..."                 
        except Exception as e:
            dbResult.isSuccess,dbResult.message,dbResult.columns,dbResult.rows,dbResult.rowcount = False,f"Exception : {e}",None,None,0
        finally:
            if (dbResult.connected):
                LoCursor.close()                
            self.closeConn(LoConn)        
        return dbResult

    def CallFun(self, AsName :str, AoParams :list, AoType :type) -> ardbResult:        
        dbResult = ardbResult()
        try:            
            dbResult.message = "Open database connection"
            LoConn , dbResult.connected = self.openConn()
            if (dbResult.connected):
                dbResult.message = "Open cursor"
                LoCursor = LoConn.cursor()                
                dbResult.message = "Execute Data"
                dbResult.data = LoCursor.callfunc(AsName , AoType, AoParams)                
                dbResult.isSuccess,dbResult.message = True, "Function Exceuted ..."                 
        except Exception as e:
            dbResult.isSuccess,dbResult.message,dbResult.columns,dbResult.rows,dbResult.rowcount = False,f"Exception : {e}",None,None,0
        finally:
            if (dbResult.connected):
                LoCursor.close()                
            self.closeConn(LoConn)        
        return dbResult


from setuptools import setup, find_packages

setup(
    name='AbdallahRadwanLib',
    version='1.5',    
    description='My First Package for Python Projects with PyPI Package - October 2024',  
    # package_dir={"": "app"},
    # packages=find_packages(where="app"),
    packages=find_packages(),
    include_package_data=True,  # Includes non-code files if specified in MANIFEST.in
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/abdorado1984/AbdallahPackage',
    author='Abdallah Radwan',
    author_email='AbdallahRadwan2011@gmail.com',    
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",        
    ],    
    python_requires='>=3.12',   
)

# packages=['arUtilities'],
